package spApp;

use strict;
use base 'CGI::Application';

use HTML::Template;
use CGI::Session;
use Apache::DBI;
#use DBI;
use Data::Dumper;

##################################
### FUNCTIONS ###
##################################

sub redirect_pagename {
	my $self = shift;
	my $pagename = shift;
	
	$self->redirect_runmode(undef, undef, undef, undef, undef, $pagename); #arent you glad we at least dont have to do nasty calls like this all over the place?
}

sub redirect_runmode {
	my $self = shift;
	my $runmode = shift;
	my $valuesearch_reset = shift; #if we get this, pass it forward to force a reset of the valuesearch form
	my $retry_form = shift; #if we get this, pass it forward to force any generated editform to be retried. (added so that we can return directly to the client signup screen if they made a form error, bypassing the error screen altogether.)
	my $refresh_opener = shift; #if we get this then we'll return some javascript to refresh the "opener" page (this was added to get the client side main list to be refreshed from the client side login popup when a successful login is done)
	my $custom_query_params = shift; #customized stuff to add on to the query ... a hashref of name => value pairs.
	my $pagename = shift; #ok so we can redirect to a page instead of a application runmode with this.
	
	#we should check for refresh_opener in the cgi too.
	if (!$refresh_opener) {
		my $cgi = $self->query();
		if ($cgi->param('refresh_opener')) {
			$refresh_opener = 1;
		}
	}			
	
	my $session = $self->param('_session');
	my $session_id = $session->id();
	
	#include the new interaction ID __AND__ the session id in all redirects so resulting screens can find their params.
	my $url_str;
	if ($runmode) {
		$url_str = $self->param('_script_url_prefix') . $self->param('_script_name') . "?rm=" . $runmode . "&interaction_id=" . $self->param('new_interaction_id')  . "&session_id=$session_id";
	} elsif ($pagename) {
		$url_str = $pagename;
	}
			
	if ($valuesearch_reset) {
		#also includee the searchform reset directive if required
		$url_str .= '&valuesearch_searchform_reset=1';
	}
	
	if ($retry_form) {
		#also includee the editform retry (make it fill in whatever values were filled in last by the user)
		$url_str .= '&retry_form=1';
	}

	if ($refresh_opener) {
		$url_str .= '&refresh_opener=1'; #this is a better way that the execute-immediately-and-fuck-up-the-session client side javascript that you see commented out below. This will just cause similar javascript to be executed when the onLoad event completes for the redirect page.
	}

	if ($custom_query_params) {
		foreach (keys(%$custom_query_params)) {
			$url_str .= '&' . $_ . '=' . $custom_query_params->{$_};
		}
	}
	#print STDERR "redirect_runmode: going to send browser to '$url_str'\n";
		
	$self->header_add(-location => $url_str);
	$self->header_type('redirect');
	#return '';
		
	my $html = qq{
		<HTML>
			<HEAD>
				<TITLE>Application Redirect</TITLE>
				<meta http-equiv="Refresh" CONTENT="0;URL=$url_str">
				<meta http-equiv="pragma" content="no-cache">
				<meta http-equiv="cache-control" content="no-cache">
				<script type="text/javascript">
					window.location.href = '$url_str';
				</script>
			</HEAD>
		</HTML>
	};

	return $html;
}

#########################
### UTILITY FUNCTIONS ###
#########################
sub standard_setup {
	#sets the active application zone and picks up the debug mode toggle directive.
	my $self = shift;
	my $runmode = shift;
	my $cgi = $self->query();
	my $session = $self->param('_session');

	#turn debug mode on/off if required. this will cause the param debug_mode to be set in the template on load_template.
	if (lc($cgi->param('debug_mode')) eq 'on') {
		$session->param('debug_mode', 1);
	} elsif (lc($cgi->param('debug_mode')) eq 'off') {
		$session->param('debug_mode', 0);
	}
	$self->param('debug_mode', $session->param('debug_mode')); #and set an app param with whatever the debug mode status is (makes for easy mode detection without haveing to ref the session)

	#print STDERR "standard_setup (of runmode " . $self->param('runmode') . "):  here1 with: " . Dumper($session->param('client_conf_params')) . "\n";

}

#Note: this sub stolen from sitePilot::Application::Admin, though default params sent to template are different now
sub load_tmpl {
	# grab arguments
	my $self	= shift;
	my $tmpl_file	= shift;

	# create html template object
	my $t = HTML::Template->new_file
	(
		$tmpl_file,
		die_on_bad_params	=> 0,
		loop_context_vars	=> 1,
		global_vars		=> 1,
		path			=>
		[
			$self->param('tmpl_path'),
			$ENV{DOCUMENT_ROOT} #so we can use sitepilot pages as templates! w00p.
		],
		search_path_on_include	=> 1,
		max_includes 		=> 10000
	);

	# set basic app file parameters
	#$t->param('somestandardparam' => 'thevalue');
	$t->param( 'APP_LABEL'	=> $self->param('_app_label') );
	$t->param( 'APP_NAME'	=> $self->param('_app_name') );
	$t->param( 'SCRIPT_NAME'=> $self->param('_script_url_prefix') . $self->param('_script_name') );
	$t->param( 'SCRIPT_URL_PREFIX' => $self->param('_script_url_prefix') );
	$t->param('vhost_root' => $self->param('_vhost_root') );
	$t->param('mail_return_name' => $self->param('_mail_return_name') );

	#my $session = $self->param('_session');
	$t->param('debug_mode' => $self->param('debug_mode'));

	return $t;
}

sub render_interface_tmpl {
	my $self = shift;
	my $sub_tmpl = shift; #pass scalar or array ref of sub tmpl names ... each will get all tmpl_params and get rendered into the master tmpl.
	my $tmpl_params = shift; #this must be a hashref.
	
	#we're going to load the master template, and a subtemplate. we're going to make all the tmpl_params avaialble to both templates, then we're going to put the output of the subtemplate into a param in the master template (the param will be named similar to the sub tempalte name)

	#get the master, plug vars in
	my $master = $self->get_master_tmpl();
	$master->param($tmpl_params);

	if ($sub_tmpl) {
		if (ref($sub_tmpl) ne 'ARRAY') {
			$sub_tmpl = [ $sub_tmpl ]; #well it is now.
		}
		
		foreach my $sub_tmpl_name (@$sub_tmpl) {
			#load the subtmpl
			my $sub_tmpl = $self->load_tmpl($sub_tmpl_name);
			$sub_tmpl->param($tmpl_params);
		
			#get the var name to use for the subtmpl
			my $sub_tmpl_var = $sub_tmpl_name;
			$sub_tmpl_var =~ s|\.tmpl$|_tmpl|;
		
			#plug into master
			$master->param($sub_tmpl_var => $sub_tmpl->output());
		}
	}
	
	return $master->output();
}	

sub ch_debug {
	my $self = shift;
	#always pass a var (should be a var ref for dump mode) and a mode
	my $config = $self->param('_config');
	my $session = $self->param('_session');

	if ($config->{suppress_debug_output} || !$session->param('debug_mode')) {
		return;
	}

	my $var = shift;
	my $mode = shift; #either "print" or "dump", assume dump if unspecified
	
	#return immediately if no var passed
	if (!$var) { return 0;  }
	
	#assign default mode if none set
	if (!$mode) { $mode = 'dump';   }
	
	#print header if not already done
	if (!$self->param("CH_DEBUG_HEADER_SENT")) {
	       print "Content-type: text/html\n\n";
	       $self->param("CH_DEBUG_HEADER_SENT", 1);
	}
	
	if ($mode eq 'dump') {
	       print "<pre>" . Dumper($var) . "</pre>";
	} else {
	       #simpler to print for any mode other than 'dump' .. easy
	       print $var;
	}
	
	print "<br><br>\n\n";
	
	return 1;
}

sub get_now {
	my $self = shift;

	#build the date because we need to include all events from today, not just those starting after NOW() which will have already occurred by the time the faclitator is ready to take attendance!
	my ($year, $month, $day) = (localtime)[5,4,3];
	$year += 1900;
	$month = sprintf("%02d", $month+1); #the month is 0 based.
	$day = sprintf("%02d", $day);
	return "$year$month$day";
}
	
sub send_email {
	#this is an old version for spiserver using sendmail. most recently i've added a Mailer.pm which does it via smtp and does other things different.

	my $self = shift;
	my $message_params = shift;
	my $message_template = $message_params->{template};
	my $template_params = $message_params->{template_params};

	#check to make sure there is a real email address provided before proceeding (with qmail we wouldnt care, but with sendmail having a bad email address gives errors.)
	my $email_provided = 0;
	my @parts = split(/\s/, $message_params->{recipient_formatted});
	my $email_part = $parts[0];
	if ($email_part =~ /^[\w.-]+\@([\w-]+\.)+\w+$/) {
		$email_provided = 1;
	}
	
	if ($email_provided) {
		#print STDERR "\ngot an email addr, doing the mailout";
		my $config = $self->param('_config');
	
		#the activation email we are sending will welcome the new user to the system and include the activation code inside a link to the account activation mode. (see notes inside there for more info)
		my $t;
		#if ($message_params->{client_type} eq 'offshore') {
		#	$t = $self->load_tmpl('mail_ae_master.tmpl');
		#} else {
		#	#this is what would happen most of the time.
		#	$t = $self->load_tmpl('mail_jc_master.tmpl');
		#}
		$t = $self->load_tmpl('mail_master.tmpl');
		
		my $mail_from    = $config->{mail_return_address} . " (" . $config->{mail_return_name} . ")";
		my $mail_subject = $message_params->{subject};
		
		#if ($message_params->{client_type}) {
		#	my $company_names = { 'other' => 'JC Clark', 'offshore' => 'Alexander Ernst'};
		#	my $company_sites = { 'other' => 'jc.spiserver.com', 'offshore' => 'www.alexanderernst.com'};
		#	$template_params->{company_name} = $company_names->{$message_params->{client_type}};
		#	$template_params->{company_site} = $company_sites->{$message_params->{client_type}};
		#	
		#	#override mail_from if a client type was specified.:
		#	$mail_from    = $config->{mail_return_address} . " (" . $company_names->{$message_params->{client_type}} . ")";
		#	#messages sent with a client type will need the company name prefixed to the subject
		#	$mail_subject = $company_names->{$message_params->{client_type}} . $mail_subject;
		#}
		
		#replace the recipient specific information in the message template
		$template_params->{'lang_' . $self->param('lang')} = 1;
		$t->param($template_params);
		$t->param('host_name' => $config->{mail_resources_host});
		$t->param('message_title' => $message_params->{subject}); #just what we want in the title tag (not even really used for an html email).
		
		my $message_contents = $t->output();
		
		# create mime template for full html email
		require MIME::Lite;
		my $msg = MIME::Lite->build
		(
			Type	=> 'TEXT/HTML',
			Data	=> $message_contents,
		);
		
	
		$msg->replace('X-Mailer' => $self->param( '_app_label'));
		$msg->add("Return-Path"  => $config->{mail_return_address} );
		$msg->add("Errors-To"	 => $config->{mail_return_address} );
		$msg->add("X-Errors-To"	 => $config->{mail_return_address} );
		$msg->add('From'	 => $mail_from);
		$msg->add('To'		 => $message_params->{recipient_formatted}); #ex: $attendee_display_values->{email} . " ($attendee_name)"
		$msg->add('Subject'	 => $mail_subject);
		
		# get the content of the mime
		my $mime	= $msg->as_string();
		
		#and send the message on its merry way.
		open (SENDMAIL, '| /usr/lib/sendmail -t -i');
		 print SENDMAIL $mime;
		close(SENDMAIL);
	}
	#else {
	#	print STDERR "\n no email given, no mailout happenin'";
	#}
}


sub setup_session {
	my $self = shift;
	my $cgi = $self->query();
	my $dbh = $self->get_dbh();

	#the session object will find the session id (defualt param name is CGISESSID) in the $cgi object
	#also, for driver:MySQL the default tablename is "sessions", and that is wehat we'll use.
	my $session_name = 'spApp_session_id';
	CGI::Session->name($session_name);
	my $session_id = $cgi->cookie($session_name) || $cgi->param('session_id') || undef;
	
	#going to try filesystem based sessions out to see if they alleviate the problem of the session data becoming unretreivable when the interaction count gets too high (seems to be arbitrary as to when it starts failing)
	#on spiserver2, for some unknown reason the mysql sessions just dont work. it just doesnt create the session record. I have no idea why, I spent 4 hours tyring to make it work and its just not worth it. Oh how I wish I could have been working under a mentor or someone with real deep knowledge that could have tipped me off in the right direction to solve the problem. I verified that the db worked by manually writing a record to the table with the same dbh that gets passed in here. I get no errors from creating the session it just does not work. So I cheap out and use file system sessions.
	#my $session = new CGI::Session("driver:MySQL;serializer:Storable", $session_id, {Handle=>$dbh});
	#my $session = new CGI::Session("driver:MySQL;serializer:FreezeThaw", $session_id, {Handle=>$dbh});
	if (!-e $self->param('_session_path')) {
		die "setup_session: fatal error, sessions path '" . $self->param('_session_path') . "' does not exist";
	}
	my $session = new CGI::Session("driver:File;serializer:Storable", $session_id, {Directory=>$self->param('_session_path')});
	if (!$session) {
		die "Couldnt create session";
	} else {
		#$self->ch_debug(['I think I created the session. it looks like this:', $session, 'session creation errors (if any):', $session->error()]);
	}
	
	#my $session = new CGI::Session("driver:File", $session_id, {Directory=>$self->param('_session_path')});
	$self->param('_session_id', $session->id()); #add session id here too.
	
	#set expireation rules for login_status param as well as entire session (testing with 1 minute/2minutes for now)
	my $config = $self->param('_config');
	#$self->ch_debug(['setup_session: config looks like', $config, 'browser sent this for session id: ', $session_id]);
	#print STDERR "browser sent this for session id: $session_id \n";
	$session->expire('_logged_in', $config->{session_login_expire});
	$session->expire($config->{session_expire});

	$self->param('_session', $session); #and store the session object itself.

	#we also need to make sure the cookie gets sent to the browser
	my $cookie = $cgi->cookie(
		-name => $session->name(), 
		-value => $session->id(),
	);
	$self->header_add(-cookie => [$cookie]); 
	$self->header_add(-type => 'text/html; charset=utf-8');
}

sub cgiapp_init {
	my $self = shift;


	#die $self->param('_app_name');
	#set paths used by overridden load_tmpl method
	my $vhost_root = $ENV{DOCUMENT_ROOT};
	$vhost_root =~ s|(.*)/.*|$1|; #strip last dir off (usually httpdocs) to get siteroot dir

	#this is where we would load things from a config file
		# - database connection information for use by sub connect_dbh
		# - application settings like app name
	use spApp::FlatFile;
	my $config_file = $self->param('conf') . '/conf/' . $self->param('_app_name') . '.conf';
	if (!-e $config_file) {
		die ("config file: $config_file cannot be read");
	}

	my %config = configAccessRead($config_file);
	$self->param('_config', \%config);

	my $sessions_dir;
	if ($config{sessions_dir}) {
		$sessions_dir = $config{sessions_dir};
	} else {
		$sessions_dir = 'sessions';
	}
	my $script_name = $ENV{REDIRECT_URL} ? $ENV{REDIRECT_URL} : $ENV{SCRIPT_NAME}; #if the script name part of the url is rewritten by mod_rewrite, make use of that name.
	my $web_prot = ($ENV{SERVER_PORT} eq '443') ? 'https://' : 'http://'; #decide whether we are using http or https ... hint, if the doc root is httpsdocs and we got here on port 443, we're doing https.
	$self->param( '_app_label', $config{'app_label'});	
	$self->param('_mail_subject_prefix', $config{mail_subject_prefix});
	$self->param('_vhost_root', $vhost_root);
	$self->param('tmpl_path', $self->param('tmpl_path') . '/tmpl_' . $self->param('_app_name'));
	$self->param('_script_name', $script_name); 
	$self->param('_report_output_path', $ENV{DOCUMENT_ROOT} . '/reports'); 
	$self->param('_script_url_prefix', $web_prot . $ENV{HTTP_HOST}); 
	$self->param('_htmldoc_resources', $web_prot . $config{'htmldoc_resources_host'}); 
	$self->param('_mail_return_name', $config{'mail_return_name'}); 
	$self->param('_session_path', $self->param('conf') . '/' . $sessions_dir); #added to play around with filesystem session storage (CGI::Session is not cleaning up its mess in the db so I want to see if it will clean up its mess in the fs instead)

	$self->setup_session(); #make the session object available via $self->param('_session'); NOTE: this needs access to the database which is why we read the config file first above.

	#$self->ch_debug(\%ENV);
}

#all we do here is disconnect from the db. nice and easy.
sub teardown {
	my $self = shift;

	#store specific information about the runmode call in the session at this point so that it could be available on the next call
	#this is important because we need to be able to tell if the user has moved between different application zones, thus changing certain behaviours.
	#of note is that we will want to destroy searchform_results_params if user was in a zone other than that specified in the searchform_def.
	
	#determine the application zone from the runmode (by being a smarty/silly/lazypants) and extracting it from the runmode name
	my $cgi = $self->query();
	my $session_obj = $self->param('_session');
	
	#close the database connection. There is a case when we fork that the parent process closes the dbh connection early to prevent the child process from losing its dbh when the parent exits (closing it early forces the child to get a brand new dbh). consequently, when we do this early db close we'll skip closing it here because the lines below would cause another db connectino to be made and then immediately closed which would just add unneccessary overhead.
	if (!$self->param('no_dbh_disconnect')) {
		my $dbh = $self->get_dbh();
		$dbh->disconnect if( $dbh );
	}
}

#sub get_form_gen {
#	my $self = shift;
#	if (!$self->param('_form_gen')) {
#		require formGen;
#		my $form_gen = formGen->new($self);
#		$self->param('_form_gen', $form_gen);
#	}
#	return $self->param('_form_gen');
#}

sub get_dbh {
        my $self = shift;
	#$self->ch_debug(['get_dbh']);
        $self->connect_dbh() unless( $self->param('_dbh') );
        return $self->param('_dbh');
}

# connect to the database
sub connect_dbh {
        my $self        = shift;

	#$self->ch_debug(['connect_dbh']);
	my $config = $self->param('_config');

        my $c_database  = $config->{db_name};
        my $c_server    = "DBI:mysql:$c_database:" . $config->{db_host};#also contains the port number
        my $c_username  = $config->{db_user};
        my $c_password  = $config->{db_pass};
	#die (Dumper($config));
        my $dbh         = DBI->connect( $c_server, $c_username, $c_password, {AutoCommit => 1}) || die "Can not connect to Database";
	#$self->ch_debug(['connect_dbh, dbh looks like:', $dbh]);

        $self->param( '_dbh' => $dbh );
}


1;
