package spApp::FlatFile;

##+ required files +###########################################################
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(tableAccessRead   tableAccessWrite   configAccessRead   configAccessWrite
	     getDbHeaderLine   availableUnique    checkForDuplicate);

use strict;
use Fcntl qw(:DEFAULT :flock);

###############################################################################
sub tableAccessRead
{
	# grab arguments
	my $location = shift;
	my $request  = shift;

	# declare local variables
	my %hash;
	local *DATABASE;

	# open the db and lock
	open(DATABASE, $location) or return %hash;
	flock(DATABASE, LOCK_SH) unless ($ENV{'SERVER_SOFTWARE'} =~ m|^SAMBAR.*?$|);

	# get header line
	my $header = <DATABASE>;
	chomp($header);

	# split header line into fields
	my @fields = splitHeaderLineIntoFields($header);
	my $count = @fields;

	# read the rest of the db line by line
	while(<DATABASE>)
	{
		chomp;
		my @data = split(/\t/);

		next unless(@data);

		# set wholeLine value
		$hash{$data[0]}{wholeLine} = $_;

		# set each value in using array of header fields
		for(my $x=1; $x < $count; $x++)
		{
			# strip trailing and leading white space
			$data[$x] =~ s/\s*$//;
			$data[$x] =~ s/^\s*//;

			# read line to hash
			$hash{$data[0]}{$fields[$x]} = $data[$x];
		}
	}

	# save the file header if requested
	if($request)
	{
		my $fields	= \@fields;
		$hash{header}	= $fields;
	}

	# close the db and unlock
	close(DATABASE);

	return %hash;
}



###############################################################################
sub tableAccessWrite
{
	# grab arguments
	my($hash, $location) = @_;

	# declare local variables
	local *DAT;

	# get header line from existing db
	my $header = getDbHeaderLine($location);
	return if(!$header);

	my @header = splitHeaderLineIntoFields($header);

	# remove id field from header array
	shift(@header);

	# remove header line from input hash
	delete $$hash{0};
	delete $$hash{header};

	# open db and lock for writing
	open(DAT, ">$location") or return;
	flock(DAT, LOCK_EX) unless ($ENV{'SERVER_SOFTWARE'} =~ m|^SAMBAR.*?$|);

	# print header line
	print DAT "$header\n";

	# now loop through %hash and print all data in order of @header fields
	foreach my $hashKey ( keys %$hash )
	{
		next if(!$hashKey);

		# strip leading and trailing white space
		$hashKey =~ s/\s+$//;
		$hashKey =~ s/^\s+//;

		print DAT "$hashKey";
		foreach my $hashField ( @header )
		{
			print DAT "\t";

			# strip leading and trailing white space
			$hashField =~ s/\s+$//;
			$hashField =~ s/^\s+//;

			next if(!$hashField);
			next if(!$$hash{$hashKey}{$hashField});

			print DAT $$hash{$hashKey}{$hashField};
		}

		print DAT "\n";
	}

	# close db and unlock
	close(DAT);
}

sub splitHeaderLineIntoFields
{
	# grab argument
	my $header = shift;

	# remove trailing newline
	chomp($header);

	# split header line into fields
	my @fields = split(/\t/, $header);
	foreach my $field ( @fields )
	{
		$field =~ s/\s*$//;
		$field =~ s/^\s*//;
	}

	return @fields;
}


sub getDbHeaderLine
{
	# grab arguments
	my $location = shift;

	# declare local variables
	local *DB;

	# get header line from existing db
	open DB, $location or return;

	 flock(DB, LOCK_SH) unless ($ENV{'SERVER_SOFTWARE'} =~ m|^SAMBAR.*?$|);
	 my $header = <DB>;

	close DB;

	chomp($header);
	return $header;
}

###############################################################################
sub configAccessRead
{
	# grab arguments
	my $location = shift;

	# declare local variables
	my %config;

	# open file and lock for reading
	open(CONFIG, "$location") or return %config;

	#die " got here to configAccessRead\n";
	flock(CONFIG, LOCK_SH) unless ($ENV{'SERVER_SOFTWARE'} =~ m|^SAMBAR.*?$|);
	#die " got here to configAccessRead\n";

	# read file line by line and add to config hash
	while(<CONFIG>)
	{
		chomp;
		s/^#.*//;
		s/^\[.*//;
		s/^\s*//;
		s/\s*$//;
		next unless length;
		my ($var, $value) = split(/\s*=\s*/, $_, 2);
		$config{$var} = $value;
	}

	# close file and remove lock
	close(CONFIG);

	# return hash of name=value pairs
	return %config;
}


###############################################################################
sub configAccessWrite
{
	# grab arguments
	my( $hash, $location ) = @_;

	# open file and lock for writing
	open(CONFIG, ">$location") or return;
	flock(CONFIG, LOCK_EX) unless ($ENV{'SERVER_SOFTWARE'} =~ m|^SAMBAR.*?$|);

  	# write each entry in $hash to the file
  	foreach my $entry ( keys %$hash )
  	{
   		next unless($entry);
   		print CONFIG "$entry = $$hash{$entry}\n";
  	}

	# close the file and remove lock
	close(CONFIG);
}


###############################################################################
sub replaceVars
{
	my($code, $hash) = @_;

	my($key);

	foreach $key (keys %$hash)
	{
		$code =~ s/\[\$$key\$\]/$$hash{$key}/gi;
	}

 	return $code;
}


##[ available unique ]#########################################################
# takes all the keys from a database, and finds the first available sequential
# numberic unique number
sub availableUnique
{
	my $hash = shift;

	# declare local variables
	my ($id, $newId);

	# loop through the sorted hash list
	foreach $id ( sort { $a <=> $b } keys %$hash )
	{
		next if($id =~ /[a-zA-Z]/);

		$newId++;

		return $newId if($id != $newId);
	}

	# returns the new id
	return ++$newId;
}


##[ check for duplicate ]######################################################
# takes a hash of hashes, the name of the field to compare, and a value to
# compare it to, and returns the id number from the hash of the first entry
# found that matches the value provided in the field provided.
sub checkForDuplicate
{
	my($hash, $field, $value) = @_;

	# loop through hash if the value is a string
	if($value & ~$value)
	{
		foreach my $key ( keys %$hash )
		{
			return $key if($$hash{$key}{$field} eq $value)
		}
	}

  	# loop through hash if the value is a number
  	else
  	{
		foreach my $key ( keys %$hash )
		{
			return $key if($$hash{$key}{$field} == $value)
		}
	}
}

# return true
1;