package spApp::FunMoney;

use strict;
use base 'spApp';

use Data::Dumper;

################################
### SUBCLASS SETUP FUNCTIONS ###
################################
sub setup {
	my $self = shift;
	$self->start_mode('show_main_menu');
	$self->mode_param('rm');
	$self->run_modes(
		'show_main_menu' => 'show_main_menu',

		'show_login' => 'show_login',
		'login' => 'login',
		'logout' => 'logout',

		'show_profiles_list' => 'show_profiles_list',
		'show_profile' => 'show_profile', #gotta get this goin with the totalling up and all that. and when this stuff puts a load on, then its time to start trusting the str itself to maintain its version of things and revalidate the records periodically instead of every. single. pageview.

		'show_create_profile' => 'show_create_profile',
		'create_profile' => 'create_profile',

#		'show_find_profiles' => 'show_find_profiles', #Not Yet Implemented (NYI)
#		'find_profiles' => 'find_profiles', #NYI
		
		'show_post_offer' => 'show_post_offer',
		'post_offer'      => 'post_offer',

		'show_accept_offer' => 'show_accept_offer',
		'accept_offer'      => 'accept_offer',

	);
	
}

sub show_login {
	my $self = shift;
	my $tmpl_params = shift;
	
	my $cgi = $self->query();
	my $md5 = $cgi->param('md5') ? $cgi->param('md5') : $tmpl_params->{md5};

	if (!$md5) {
		#if not, forget it
		return $self->show_profiles_list({form_error => 1, form_error_text => 'wtf mate - need an md5'});
	}
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);
	my $profile_row = $prof_man->get_validated_profile_row($md5);
	if (!$profile_row) {
		$self->ch_debug(['hey mofo', $self->param('validate_profile_error')]);
		return $self->show_profiles_list({form_error => 1, form_error_text => 'Profile not found (sorry)'});
	}
	$tmpl_params->{chosen_profile_name} = $profile_row->{name};
	$tmpl_params->{md5}                 = $profile_row->{md5};
		
	$self->ch_debug(['show_login: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('login.tmpl', $tmpl_params);
}

sub login {
	my $self = shift;
	my $cgi = $self->query();
	my $md5 = $cgi->param('md5');
	my $passphrase = $cgi->param('passphrase');

	my $error = 0;
	my $errortext = '';
	
	if (!$passphrase) {
		$error = 1;
		$errortext .= 'Did not recieve passphrase from the cgi. ';
	}
	if (!$md5) {
		$error = 1;
		$errortext .= 'Did not get an md5 from the cgi. ';
	}

	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);
	my $row = $prof_man->get_validated_profile_row($md5);

	#does passphrase match?
	my $pf_md5 = $prof_man->calc_md5([$passphrase]);
	if ($pf_md5 ne $row->{pf_md5}) {
		$self->ch_debug(['login: $pf_md5 and $row->{pf_md5} dont match', $pf_md5, $row->{pf_md5}]);
		$error = 1;
		$errortext .= 'Passphrase incorrect. ';
	}

	if ($error) {
		return $self->show_login({form_error => 1, form_error_text => $errortext});
	}
	
	my $session = $self->param('_session');
	$session->param('user_profilerow' => $row);
	$session->param('user_passphrase' => $passphrase);
	$session->param('user_md5'        => $row->{md5});
	$session->param('active_profile'  => 1);

	return $self->show_main_menu();
}

sub logout {
	my $self = shift;
	my $session = $self->param('_session');
	$session->param('user_profilerow' => undef);
	$session->param('user_passphrase' => undef);
	$session->param('user_md5'        => undef);
	$session->param('active_profile'  => undef);
	
	return $self->show_main_menu();
}

sub get_active_profile_row {
	my $self = shift;

	my $session = $self->param('_session');
	my $row = $session->param('user_profilerow');
	if (!$row) { return undef; }
	$row->{active_profile} = 1;
	return $row;
}

sub show_profiles_list {
	my $self = shift;
	my $tmpl_params = shift;

	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);
	my $list = $prof_man->find_profiles({});
	$tmpl_params->{profile_list} = $list;	

	$self->ch_debug(['show_profiles_list: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('profiles_list.tmpl', $tmpl_params);
}
	
sub show_main_menu {
	my $self = shift;
	my $tmpl_params = shift;

	$tmpl_params->{main_menu} = 1;
	$self->ch_debug(['show_main_menu: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('main_menu.tmpl', $tmpl_params);
}

sub show_create_profile {
	my $self = shift;
	my $tmpl_params = shift;

#	my $cgi = $self->query();

	$self->ch_debug(['show_create_profile: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('create_profile.tmpl', $tmpl_params);
}	

sub create_profile {
	my $self = shift;

	my $cgi = $self->query();
	my $password = $cgi->param('password');
	my $email    = $cgi->param('email');
	my $name     = $cgi->param('name');
	my $blurb    = $cgi->param('blurb');
	my $store_private_key = $cgi->param('store_private_key');
	my $errortext = '';
	
	my $error = 0;
	if (!$email || !$name || !$password) {
		$error = 1;
		$errortext .= 'Missing email, name, or passphrase. ';
	}
	if ($email !~ /^[\w.-]+\@([\w-]+\.)+\w+$/) {
		$error = 1;
		$errortext .= 'Email is stupidly formatted. Come on. ';
	}
	if ($password ne 	$cgi->param('password2')) {
		$error = 1;
		$errortext .= 'Passphrases dont match. Please try again. ';
	}
	
	if ($error) { 
		$self->ch_debug(['create_profile: had input errors, sending back to show_create_profile']);
		return $self->show_create_profile({ 
			form_error => 1,
			form_error_text => $errortext, 
			email => $email, 
			name => $name,
			blurb => $blurb,
			
		});
	 }
	#the above will actually come from db as the below will actually happen after an activate.

	#being here means we got all the input we need.
		#so we'll generate the keys and save them. right now we're going to store the private key by default so that people can use this website to transact. 
		#soon enough, I'd like to be able to NOT keep private keys around (even though they're encrypted with the profile-owners' passphrase) and allow profile-owners to black box their own crypto.
	
	require spApp::ProfileManager;

	$self->ch_debug(['the email was : ', $email]);

	my $prof_man = spApp::ProfileManager->new($self);

	#meat
	my $profile_md5 = $prof_man->create_profile($password, $email, $name, $blurb, $store_private_key);

	my $result = "Your profile hash is: $profile_md5 <br> Remember your passphrase. <br> Have some fun :)";
	return $self->show_main_menu({main_menu_message => $result });
}


#NYI
#sub show_find_profiles {
#	my $self = shift;
#	my $tmpl_params = shift;
#
#	return $self->output_master('find_profiles.tmpl', $tmpl_params);
##	my $cgi = $self->query();
#
#	$self->ch_debug(['show_find_profiles: with these tmpl_params:', $tmpl_params]);
#}

sub show_post_offer {
	my $self = shift;
	my $tmpl_params = shift;

	my $session = $self->param('_session');
	if ($session->param('active_profile')) {
		$tmpl_params->{active_profile} = 1;
		$tmpl_params->{from_md5} = $session->param('user_md5');
	}
	
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);
	my $list = $prof_man->find_profiles({
		preselect => $tmpl_params->{to_md5},
		omit      => $tmpl_params->{from_md5},
	});
	$tmpl_params->{profile_list} = $list;	

	$self->ch_debug(['show_post_offer: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('post_offer.tmpl', $tmpl_params);
}	

sub post_offer {
	my $self = shift;
	my $cgi = $self->query();
	my $from_md5      = $cgi->param('from_md5');
	my $passphrase    = $cgi->param('passphrase');
	my $amount        = $self->like_mysql_decimal($cgi->param('amount'), 2); #hopefully doing this at this time staves off any other problems related to number formatting mismatches!!
	my $currency_unit = $cgi->param('currency_unit');
	my $desc          = $cgi->param('desc');
	my $to_md5        = $cgi->param('to_md5');
	
	#$self->ch_debug(['I formatted the amount field to: ', $amount, ' now will die until that is right!']);

	my $session = $self->param('_session');
	if ($session->param('active_profile')) {
		$passphrase = $session->param('user_passphrase');
		$from_md5   = $session->param('user_md5');
		$self->ch_debug(['post_offer with logged in user md5 of', $from_md5]);
	}

	my $errortext = '';
	my $error = 0;
	if (!$from_md5 || !$passphrase || !$amount || !$currency_unit || !$to_md5) {
		$errortext .= 'Missing from_md5, passphrase, amount, currency unit, or to_md5. ';
		$error = 1;
	}
	
	#other error conditions are that the to_md5 or from_md5 could be illegitimate. they have to be checked.
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);

	my $from_checks_out = $prof_man->validate_profile($from_md5);
	if (!$from_checks_out) {
		$errortext .= 'From md5 failed to validate: ' . $self->param('validate_profile_error') . '. ';
		$error = 1 
	};

	my $to_checks_out = $prof_man->validate_profile($to_md5);
	if (!$to_checks_out) {
		$errortext .= 'To md5 failed to validate: ' . $self->param('validate_profile_error') . '. ';
		$error = 1 
	};

	my $offer_id;
	if (!$error) {
		#ok we can do it. we can create the signed offer record. what happens if the passphrase is wrong?
		require spApp::TxManager;
		my $txm = spApp::TxManager->new($self);

		#meat
		$offer_id = $txm->post_offer({
			from_md5 => $from_md5,
			passphrase => $passphrase, 
			amount => $amount, 
			currency_unit => $currency_unit, 
			desc => $desc ? $desc : "None Given", 
			to_md5 => $to_md5,
		});
		
		if (!defined($offer_id)) {
			$error = 1;
			$errortext = $self->param('post_offer_error');
		}
	}
	
	if ($error) { 
		$self->ch_debug(['post_offer: had input errors, sending back to show_post_offer']);
		return $self->show_post_offer({ 
			form_error => 1,
			form_error_text => $errortext,
			from_md5 => $from_md5,
			passphrase => $passphrase, 
			amount => $amount, 
			currency_unit => $currency_unit, 
			desc => $desc, 
			to_md5 => $to_md5,
		});
	}


	my $result;
	#of course with the new !defined check above we will probably always return a success if we got here.
	if (!defined($offer_id)) {
		$result = "Error posting offer: " . $self->param('post_offer_error');
	} else {
		$result = "Successfully posted the offer. Record id $offer_id. <h6>Note: Re-POSTING the request again (by, say, pressing F5) will result in an ADDITIONAL offer being entered, which can then also be accepted (thus affecting your cumulative balance twice). Doing that would be silly, but if two people are going to use the same computer to post / accept, then umm, maybe close browser window in between, or something?</h6>";		
	}		
		
	return $self->show_main_menu({main_menu_message => $result });  
}

sub show_accept_offer {
	my $self = shift;
	my $tmpl_params = shift;

	my $session = $self->param('_session');
	if ($session->param('active_profile')) {
		$tmpl_params->{active_profile} = 1;
	}

	require spApp::TxManager;
	my $txm = spApp::TxManager->new($self);
	my $list = $txm->find_offers({
		to_md5    => $session->param('user_md5'),
		include_name_email_of => 'from_md5',
	});
	$tmpl_params->{offer_list} = $list;
	if (scalar(@$list)) { $tmpl_params->{offer_list_hasitems} = 1; }



	$self->ch_debug(['show_accept_offer: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('accept_offer.tmpl', $tmpl_params);
}	

sub accept_offer {
	my $self = shift;
	my $cgi = $self->query();
	my $to_md5     = $cgi->param('to_md5');
	my $passphrase = $cgi->param('passphrase');
	my @offer_ids   = $cgi->param('offer_id');
	my $offer_id = $offer_ids[0]; #ok actually will loop over the list below. should work. #only going to deal with one right now ... just because of the possible error conditions.
	
	my $session = $self->param('_session');
	if ($session->param('active_profile')) {
		$passphrase = $session->param('user_passphrase');
		$to_md5     = $session->param('user_md5');
		$self->ch_debug(['accept_offer with logged in user md5 of', $to_md5, 'offer id infos', \@offer_ids, $offer_id]);
	}

	my $error = 0;
	my $errortext = '';
	if (!$to_md5 || !$passphrase || !$offer_id) {
		$errortext .= 'Missing to_md5, passphrase, or offer id. ';
		$error = 1;
	}
	
	#other error conditions are that the to_md5 or from_md5 could be illegitimate. they have to be checked.
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self);

	my $to_checks_out = $prof_man->validate_profile($to_md5);
	if (!$to_checks_out) {
		$errortext .= 'Profile Md5 validate error: ' . $self->param('validate_profile_error') . '. ';
		$error = 1 
	};

	#ok we at least got the right input. now we can try to do the actual acceptance.
	my $accept_result;
	if (!$error) {
		require spApp::TxManager;
		my $txm = spApp::TxManager->new($self);

		#meat
		#meat upgrade: ok do it for all the offers. umm not quite sure what we'll see if there are any errors on any of them. but any error conditions on the offers should have already happened when trying to list them (and in future we'll probably take action on discovered-bogus-offers)
		foreach (@offer_ids) {
			$accept_result = $txm->post_offer_acceptance({
				md5 => $to_md5,
				offer_id => $_,
				passphrase => $passphrase, 
			});
		}

		if (!$accept_result) {
			$error = 1;
			$errortext = $self->param('post_offer_acceptance_error');
		}
	}
	
	if ($error) { 
		$self->ch_debug(['accept_offer: had errors, sending back to show_post_offer']);
		return $self->show_accept_offer({ 
			form_error => 1,
			form_error_text => $errortext,
			to_md5 => $to_md5,
			passphrase => $passphrase, 
			offer_id => $offer_id, 
		});
	}
	
	my $result;
	#of course with the new !$accept_result check above we will probably always return a success if we got here.
	if (!$accept_result) {
		$result = "TxManager post_offer_acceptance error: " . $self->param('post_offer_acceptance_error') . '. ';
	} else {
		$result = "Successfully accepted the offer(s). (And currently should be also validating the whole offer/accept record after doing so just for fun, so seeing this means that worked out)";
	}		
		
	return $self->show_main_menu({main_menu_message => $result });
	#used to return "The acceptance has been posted - if you can see this, then SOMETHING is working right! :)";
}

sub show_profile {
	my $self = shift;
	my $tmpl_params = shift;
	
	#so we can get the md5 from the cgi
	my $cgi = $self->query();
	$tmpl_params->{md5} = $cgi->param('md5');

	#and if not that then from the active profile row. gotta be logged in to see this screen eh.
	if (!$tmpl_params->{md5}) {
		my $row = $self->get_active_profile_row();
		$tmpl_params->{md5} = $row->{md5};
	}

	#we should have a md5 now. .. lets get the data.
	if ($tmpl_params->{md5}) {
		require spApp::ProfileManager;
		my $prof_man = spApp::ProfileManager->new($self);

		#find out the profile info for the selected md5 (which could be any md5 on the system since i want people to be able to view this screen, see tx of a profile, and click the party name to go see that person's profile and so on)
		my $profile_row = $prof_man->find_profiles({md5 => $tmpl_params->{md5}});
		foreach (keys(%$profile_row)) {
			$tmpl_params->{'profile_' . $_} = $profile_row->{$_};
		}

		my $hours_balance = 0; #hrmm can we calc the balance? maybe. lets see, we can add up the amounts of all the offers we posted. we can then add up the amounds of all the offers we accepted. accepted_sum - posted_sum = balance. right?

		require spApp::TxManager;
		my $txm = spApp::TxManager->new($self);

		#find (completed) offers-posted by this person
		my $posted_list = $txm->find_offers({
			from_md5    => $tmpl_params->{md5},
			accepted  => 'IS NOT NULL',
			include_name_email_of => 'to_md5',
		});
		$tmpl_params->{posted_offer_list} = $posted_list;
		if (scalar(@$posted_list)) { 
			my $currency_totals = $txm->sum_amounts({offer_list => $posted_list});
			$hours_balance -= $currency_totals->{HOURS};
			$tmpl_params->{posted_offer_list_hasitems} = 1; 
		}
		
		#find (completed) offers-accepted by this person
		my $accepted_list = $txm->find_offers({
			to_md5    => $tmpl_params->{md5},
			accepted  => 'IS NOT NULL',
			include_name_email_of => 'from_md5',
		});
		$tmpl_params->{accepted_offer_list} = $accepted_list;
		if (scalar(@$accepted_list)) {
			my $currency_totals = $txm->sum_amounts({offer_list => $accepted_list});
			$hours_balance += $currency_totals->{HOURS};
			$tmpl_params->{accepted_offer_list_hasitems} = 1; 
		}
		
		$tmpl_params->{hours_balance} = $hours_balance;

	}	
	
	$self->ch_debug(['show_profile: with these tmpl_params:', $tmpl_params]);
	return $self->output_master('profile.tmpl', $tmpl_params);

}	

#----------------------------------------------------
sub like_mysql_decimal {
	#the only point of this is so that certain numbers (today it is amount) in the offer message are saved in the same format they will be in the db row.
	my $self = shift;
	
	my $number = shift;
	my $decimals = shift;

	$number =~ s|(\.\d{$decimals})\d*$|$1|e; #truncate anything after the nth decimal point.
	$number = sprintf('%.2f', $number); #make sure it zero filled on the end.
	return $number;
}

sub cgiapp_prerun {
	my $self = shift;
	my $rm = shift;

	$self->standard_setup($rm); #sets the active application zone and picks up the debug mode toggle directive.
}

sub get_master_tmpl {
	my $self = shift;
	#this idea of using a master template for the entire application is in interesting .. will it work? I hope so! would make things smoother

	my $session = $self->param('_session');
	#at least the name of the template is defined only once! ha!
	my $t = $self->load_tmpl("funmoney_master.tmpl");
	return $t;
}

sub output_master {
	my $self = shift;
	my $sub_tmpl = shift;
	my $tmpl_params = shift;
	
	#any common stuff should get added in here.
	my $profile_row = $self->get_active_profile_row();
	if (defined($profile_row)) {
		$self->ch_debug(['output master including this name of active profile: ', $profile_row->{name}, $profile_row]);
		$tmpl_params->{active_profile} = 1;
		$tmpl_params->{active_profile_name} = $profile_row->{name};
	}

	#$self->ch_debug(['output_master: master and sub templates will get these params', $tmpl_params]);
	my $master = $self->get_master_tmpl();
	$master->param($tmpl_params);

	if (ref($sub_tmpl) ne 'ARRAY') {
		$sub_tmpl = [ $sub_tmpl ]; #now it is. haha.
	}
	
	my $st_cnt = 0;
	foreach (@$sub_tmpl) {
		$st_cnt++;
		#$self->ch_debug(['will load a sub tmpl called
		my $t = $self->load_tmpl($_);
		$t->param($tmpl_params);
		my $var_name = $_;
		$var_name =~ s|\.|_|g; #so foo.tmpl gets turned into a var foo_tmpl ... lame.html to lame_html .. ya dig?
		$master->param($var_name => $t->output());

		if (scalar(@$sub_tmpl) == 1) {
			#special allowance, if only one sub tmpl, for convenience make use of the sub_tmpl var. ... make it easy for the master template!
			$master->param('sub_tmpl' => $t->output());
		}
	}
	return $master->output();
}
	

#########################
### RUNMODE FUNCTIONS ###
#########################

sub user_logout {
	my $self = shift;

	my $session = $self->param('_session');
	$self->ch_debug('maybe i should delete the session?');

	return $self->redirect_pagename('/');
}

sub mail_preview {
	my $self = shift;

	my $config = $self->param('_config');
	my $t = $self->load_tmpl('mail_master.tmpl');
	
	$t->param('host_name' => $config->{mail_resources_host});
	$t->param('message_title' => 'Preview of Email Message Template'); #just what we want in the title tag (not even really used for an html email).

	$t->param('simple_preview', 1);
	return $t->output();
}

1;
