package spApp::ProfileManager;

use strict;
use Storable qw(freeze thaw); $Storable::canonical = 1;
use Digest::MD5;
use Crypt::RSA;

# constructor
sub new {
	my $class = shift;
	my $self = {};
	bless $self, $class;
	
	$self->{wa} = shift; #suck in a ref to the webapp for good measure. whatever.
	
	return $self;	
}

sub create_profile {
	my $self = shift;
	my ($password, $email, $name, $blurb, $store_privatekey) = @_;
	
	#generate the keys ... takes a minute.
	require spApp::RSACrypto;
	my $crypto   = spApp::RSACrypto->new($self->{wa});
	#my $keys = {private => 'scary-binary-data', private_ser => 'scary-binary-data', public => 'scary-binary-data', public_ser => 'scary-binary-data'};
	my $keys = $crypto->generate_keys($password, 1);
	my $md5 = $self->calc_md5([$name, $email, $keys->{public_ser}]); #so we're taking a md5 of the name+public_key to sort of associate them. that way if a loser admin changes someones public key or name info then the hash would not match what this person will have given out as their profile-verifier-hash.
	
	#blaargh. .. am gonna create, to allow ppl the even slightly less secure convenience of logging in, a md5 of the passphrase and store that with the profile as well. it dosnt matter its just for the ability to log in.
	my $pf_md5 = $self->calc_md5([$password]);
	
	#make the initial profile record
	my $dbh = $self->{wa}->get_dbh();
	my $sql = "INSERT INTO profile (name, email, blurb, public_key, private_key, md5, pf_md5) VALUES (?, ?, ?, ?, ?, ?, ?)";
	my @bind_params = ($name, $email, $blurb, $keys->{public_ser}, ($store_privatekey ? $keys->{private_ser} : undef), $md5, $pf_md5);
	$dbh->do($sql, undef, @bind_params) or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;
	
	my $message = 'Welcome. I need a better message than this. maybe i should take a MD5 of the public key for a more easy to manage identifier? Also, sending the private key by email is probably bad. how to communicate them? ssl? the whole interface should be ssl.';
	$message   .= "\n\nOh here is your profile md5. You use it to verify that no asshole STR admin has fucked around with your profile record (well at least, right now, for the name+public_key fields which are the most important public infos)\n\nMD5: $md5";
	
	#send an email with the keys.
   #(or not)
#	my $filename = $name; $filename =~ s|\s+|_|g;
#	require spApp::Mailer;
#	my $mailer = spApp::Mailer->new($self->{wa});
#	$mailer->send_email({
#		to      => $email,
#		to_name => $name,
#		subject => 'New Profile, Homedog',
#		text    => $message,
#		attachments => [{
#			type     =>'application/fuckmoney-cryptogram',
#			data     => $keys->{public_ser},
#			filename => $filename . '.public',
#		},{
#			type     =>'application/fuckmoney-cryptogram',
#			data     => $keys->{private_ser},
#			filename => $filename . '.private',
#		}],
#	});
	
	return $md5;
}

sub find_profiles {
	my $self = shift;
	
	#this is to return data structures that can be used for output.
	
	my $args = shift;
	my $terms = $args->{terms};
	my $preselect = $args->{preselect}; #an md5 to mark as 'selected'
	my $omit      = $args->{omit};      #an md5 to leave out of the list (because I'm logged in as him and dont want to post offers to myself that'd be stupid - at least I think it would be stupid right now).
	my $specific_md5 = $args->{md5};    #to get just a single row back.
	#my $searchability = $args->{searchability};

	my $fieldlist = "name, email, blurb, md5, CONCAT(name, ' - ', email) AS name_email";
	my $orderby   = 'name, email ASC';

	my $where = '';
	if ($terms) {
		my $where_blocks = [];
		foreach (keys(%$terms)) {
			#if ($searchability->{$_} eq 'keyword') {
				push(@$where_blocks, "($_ LIKE '%$terms->{$_}%')");
			#} 
		}
		$where = ' WHERE ' . join(' OR ', @$where_blocks);
	}
	if ($specific_md5) {
		$where = " WHERE md5 = '$specific_md5'";
	}

	my $sql = 'SELECT ' . $fieldlist . ' FROM profile' . $where . ' ORDER BY ' . $orderby;
	my $dbh = $self->{wa}->get_dbh();
	$self->{wa}->ch_debug(['gonna do this sql', $sql]);
	my $sth = $dbh->prepare($sql);
	$sth->execute() or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;
	if ($specific_md5) {
		my $row = $sth->fetchrow_hashref();
		return $row;
	}

	my $list = [];
	while (my $row = $sth->fetchrow_hashref) {
		if ($row->{md5} eq $preselect) { $row->{selected} = 1; }
		if ($row->{md5} eq $omit)      { next; }
		push(@$list, $row);
	}
	
	return $list;
}

sub get_validated_profile_row {
	my $self = shift;
	my $md5 = shift;
	return $self->validate_profile($md5, 1);
}

sub validate_profile {
	my $self = shift;
	my $md5 = shift;
	my $return_row = shift;
	
	$self->{wa}->ch_debug(['validate_profile with args: ', "md5: $md5 , return_row: $return_row"]);
	$self->param('validate_profile_error' => undef);
	
	my $fieldlist = 'id, name, email, public_key, md5';
	if ($return_row) { $fieldlist = '*'; }
	
	my $sql = 'SELECT ' . $fieldlist . ' FROM profile WHERE md5 = ?';
	my $dbh = $self->{wa}->get_dbh();
	my $sth = $dbh->prepare($sql);
	$sth->execute($md5) or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;
	my $check_existing_md5 = 0; #zero for failure. must get set to one.
	my $found_matching_md5 = 0; #zero for failure. must get set to one.
	my $good_row = undef;
	while (my $row = $sth->fetchrow_hashref()) {
		#now there should never be more than one unless some dunce edits the db and reuses the same md5 to try and fake a record. if the md5 on the row doesnt compute right, delete the row.
		$found_matching_md5 = 1; #cuz we're here so there must have been one.
		my $existing_md5 = $self->calc_md5([$row->{name}, $row->{email}, $row->{public_key}]);

		if ($existing_md5 ne $row->{md5}) {
			#ok dont crap out and delete the bad row, just skip.
			$self->param('validate_profile_error' => 'existing row has bad md5 - is someone tampering with the db?');
			$self->param('v_p_ec' => 1);
			#$row = undef;
			next;
		}
		$check_existing_md5 = 1;
		$good_row = $row;
	}
	if (!$found_matching_md5) { 
		$self->param('validate_profile_error' => 'no matching md5 found');
		$self->param('v_p_ec' => 2);
		return 0; 
	} #didnt even find a match. sorry cant help you!
	if (!$check_existing_md5) { return 0; } #found matching md5 but it did not calculate properly (someone tamplering with db?)

	if ($return_row) {
		return $good_row;
	} else {
		return 1; #getting here means we're good to go.
	}
}

sub get_profile_public_key {
	my $self = shift;
	my $md5 = shift;
	my $validate = shift; #in many contexts it will already be validated and not needed to do it again. but sometimes maybe not. so lets be flexible.

	#we could validate again too
	if ($validate) {
		if (!$self->validate_profile($md5)) { 
			$self->param('get_profile_public_key_error' => 'told to validate the md5 and it failed for some reason');
			return undef; 
		}
	}

	my $dbh = $self->{wa}->get_dbh();
	my $row = $dbh->selectrow_hashref('SELECT public_key FROM profile WHERE md5 = ?', undef, $md5);
	my $public_key_ser = $row->{public_key}; #serialized remember.
	if (!$public_key_ser) { die "get_profile_public_key: didnt even get a serialized public key back from the db - was looking for md5 '$md5'"; }
	my $public_key = thaw($public_key_ser); #now back into an object.

	return $public_key;
}

sub get_profile_private_key {
	my $self = shift;
	my $md5 = shift;
	my $passphrase = shift; #could be left undef to not do the reveal.
	my $validate = shift; #in many contexts it will already be validated and not needed to do it again. but sometimes maybe not. so lets be flexible.
	
	#we could validate again too
	if ($validate) {
		if (!$self->validate_profile($md5)) { 
			$self->param('get_profile_private_key_error' => 'told to validate the md5 and it failed for some reason');
			return undef; 
		}
	}

	my $dbh = $self->{wa}->get_dbh();
	my $row = $dbh->selectrow_hashref('SELECT private_key FROM profile WHERE md5 = ?', undef, $md5);
	my $private_key_ser = $row->{private_key}; #serialized remember.
	if (!$private_key_ser) { die "get_profile_private_key: didnt even get a serialized private key back from the db - was looking for md5 '$md5'"; }
	my $private_key = thaw($private_key_ser); #now back into an object.
	
	if ($passphrase) {
		$private_key->reveal(Password => $passphrase); #now what happens if the passphrase was wrong? -- seems to fuck up the signing function. better put an eval over there.
	}
	
	return $private_key;
}

sub calc_md5 {
	my $self = shift;
	my $fields_arr = shift; #must be an arrayref to some scalars. gonna concat them all in the order received and take a md5.

	my $ctx = Digest::MD5->new;
	foreach (@$fields_arr) {
		$ctx->add($_); 
	}
	my $md5 = $ctx->hexdigest;
	$self->{wa}->ch_debug(["the md5 of these scalars", $fields_arr, "came out to '$md5'"]);
	return $md5;
}	

sub param {
	my $self = shift;
	return $self->{wa}->param(@_);
}

1;
