package spApp::RSACrypto;
#spApp::RSACrypto object:
#	methods:
#		- encrypt
#		- decrypt
#		- sign
#		- verify
#			- think this should verify also that the message is signed by ALL parties who would have needed to sign it, recursively checking those parties to see if they have multi -sign requirements as well.
#		- generate_keys

use strict;
use Crypt::RSA;
use Storable qw(freeze thaw); $Storable::canonical = 1;

# constructor
sub new {
	my $class = shift;
	my $self = {};
	bless $self, $class;
	
	$self->{wa} = shift; #suck in a ref to the webapp for good measure. whatever.
	
	return $self;	
}

sub generate_keys {
	my $self = shift;
	my $password = shift;
	#my $serialize = shift; #bah lets do it all the time.
	#my $identity = shift; #NYI .. according to docs would be some kind of associated string like email address. no hard association currently .. we'll use our own.
	
	my $rsa = new Crypt::RSA; 
	
	my ($public, $private) = $rsa->keygen ( 
		#Identity  => 'Hammerpants <kingasshat@hammerpants.com>', #feature apparently NYI
		Size      => 1024,  
		Password  => $password, 
		Verbosity => 0,
		#Filename  => 'asshat',
	) or die $rsa->errstr();
	
	#not sure about whether the private key will actually be hidden at this point. make sure. especially since we might want to serialize it and send it via email.
	$private->hide();

	my $keys = { 'public' => $public, 'private' => $private };
	#if ($serialize) {
		$keys->{public_ser} = freeze($public);
		$keys->{private_ser} = freeze($private);
	#}
	return $keys;
	
}	

sub sign_serialized_data {
	my $self = shift;
	my $args = shift;

	my $private_key = $args->{private_key};
	my $serialized_data = $args->{serialized_data};
	
	my $rsa = new Crypt::RSA; 
	my $signature = $rsa->sign ( 
		Message    => $serialized_data, 
		Key        => $private_key
	);
		
	if (!$signature) {
		die 'RASCrypto->sign_serialized_data: failed to sign the data beacuase of the following error: ' . $rsa->errstr();
	}

	$self->{wa}->ch_debug(['sign_serialized_data: signature shaped up like:', $signature]);
	return $signature; #oooh .. the signature IS serialized its not an object.
}

sub validate_signature {
	my $self = shift;
	my $args = shift;

	my $public_key = $args->{public_key};
	my $serialized_data = $args->{serialized_data};
	my $signature = $args->{signature};

	my $rsa = new Crypt::RSA; 
	my $verify = $rsa->verify (
		Message    => $serialized_data, 
		Signature  => $signature, 
		Key        => $public_key
	);
	
	if (!$verify) {
		die 'RASCrypto->validate_signature: the signature appeared to be incorrect and invalid. error: ' . $rsa->errstr();
	}

	return $verify; #will be either 1 or undef. actually with the die statement above, we wont get here if it is undef.
	
}
	

sub whatever {
	my $self = shift;
	
	my $fail = 1; #pessimistic, no?
	
	if ($fail) { 
		return undef;
	}
	
	return 1;
}	

1;
