package spApp::TxManager;

use strict;
use Storable qw(freeze thaw); $Storable::canonical = 1;
use Time::Piece::MySQL;
use Carp;

# constructor
sub new {
	my $class = shift;
	my $self = {};
	bless $self, $class;
	
	$self->{wa} = shift; #suck in a ref to the webapp for good measure. whatever.
	
	return $self;	
}

sub post_offer {
	my $self = shift;
	my $args = shift;
	
	my $from_md5   = $args->{from_md5},
	my $passphrase = $args->{passphrase}, 
	my $amount     = $args->{amount}, 
	my $currency_unit = $args->{currency_unit}, 
	my $desc       = $args->{desc}, 
	my $to_md5     = $args->{to_md5},

	#lets sign the offer - we do this as proxy for the person since they dont have the software to do it.
		#now yes there are certainly potential problems here and why i do NOT like having the private keys here.
		#see, some unscrupulus doofus sysadmin could modify the source here to create additional offers to other party's at the same time.
		#then if the one real offer is accepted, said modified source code could then use the acceptor's passphrase to accept the other fraudulent offers created earlier by the bad sysadmin.
		#now this behaviour of course would be noticed pretty quick by going thru the transaction records and would result in the prompt removal of the evil STR from the network through widespread revokation of trust towards it by the network users. But still, i dont like the fact that said situation could easily arise which is why as soon as possible the reccomended way of posting offers should involve people using their own systems to generate the signed offer message. (and even their own no-private_key-profiles which they would communicate with any given STR before making their first offer-posting)
	$self->param('post_offer_error' => undef);
	$self->param('poec' => undef); #error by code
	
	#establish timestamp for offer
	my $time = localtime;
	my $offer_time = $time->mysql_datetime;

	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self->{wa});
	$self->{wa}->ch_debug(['post_offer with from_md5 of', $from_md5]);
	my $private_key = $prof_man->get_profile_private_key($from_md5, $passphrase); #dont bother re-validating at this point.
	if (!$private_key) { confess 'ok something went wrong didnt get a private key back from prof_man->get_profile_private_key'; }
	my $offer_message = {
		from_md5      => $from_md5,
		amount        => $amount, 
		currency_unit => $currency_unit, #yes this definitely should be turned into short form or an ID or something but for now it will just be the string 'HOURS'
		desc          => $desc, 
		to_md5        => $to_md5,
		offer_time    => $offer_time,
	};
	my $offer_message_serialzed = freeze($offer_message);

	require spApp::RSACrypto;
	my $crypto   = spApp::RSACrypto->new($self->{wa});

	#note signature will just be a block of scalar data, not an object.
	my $signature;
	my $result = eval {
		$signature = $crypto->sign_serialized_data({
			private_key     => $private_key,
			serialized_data => $offer_message_serialzed,
		});
		return 1;
	};
	$private_key->hide(); #so we dont have the passphrase floating around in memory (or so i hope)
	if (!$result) {
		$self->param('post_offer_error' => "Error signing the message. (Always seems related to bad passphrase)");
		$self->param('poec' => 1);
		return undef;
	}
		
	#ok it is signed. w00t. 
	#insert into the offers table. and get back and offer id.
	my $sql = 'INSERT INTO offer (from_md5, amount, currency_unit, `desc`, to_md5, offer_message, offer_signature, offer_time) VALUES (?, ?, ?, ?, ?, ?, ?, ?)';
	my $dbh = $self->{wa}->get_dbh();
	$dbh->do($sql, undef, ($from_md5, $amount, $currency_unit, $desc, $to_md5, $offer_message_serialzed, $signature, $offer_time)) or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;

	#bang thats it?
	#in theory now, to_md5 could visit a page and look at any offers that they have.
	return $dbh->last_insert_id(undef, undef, 'offer', 'id');
}	

sub post_offer_acceptance {
	my $self = shift;
	my $args = shift;
	
	$self->param('post_offer_acceptance_error' => undef);
	
	#so the offer is to be accepted! This is good news.
	#we need to know which offer ($offer_id), the md5 of who is accepting ($md5), and the acceptor's passphrase so we can act as proxy to sign the acceptance for them.
		#as soon as possible i want people to be able to sign their own offer acceptance messages to be validated then posted
			#what can be tampered with in this scenario what are the potential problems?
			#i dont think anything can because anyone can simply validate the signatures on the offer and the offer acceptance and if they dont check out then this transaction would be ignored as invalid (and should be reported to the network operations team).
	#we then validate the profile for the $md5, if that checks out we can act as proxy to sign the offer acceptance message (just a copy of the offer plus its signature by the offer)

	#i would like God to send a more clear message about whether I should be diverting my attention away from my SPI work to work on this stuff. Frankly I'd rather work on this stuff than on SPI stupid CMCReg (capitalist fucking scum) language translation shit.
		#i dont think asking God for clear messages will work.
		#the prospect of ditching SPI scares me because I rely on the money income (and so does Jessica to a large degress since it pays for her monthly supplement and for Harmony's daycare. Oh crap today is Friday I have to go to Jody's with bling. Maybe I'll be injured and forced to resign from SPI. i dont want to be injured. I dont want to work for money anymore either.
		#i should ask for "a scenario to occur by which i will be able to continue working on this more important stuff and not be distracted by chasing dollars and where Harmony and Jessica dont have to suffer."
	#i should not write comments to God.
	
	my $md5 = $args->{md5}, #md5 of the person who is accepting the offer ... so we can get/verify/use their profile to do the signing.
	my $offer_id = $args->{offer_id}, 

	#get the offer message and validate it.
	my $sql = 'SELECT * FROM offer WHERE id = ? AND to_md5 = ? AND accept_time IS NULL';
	my $dbh = $self->{wa}->get_dbh();
	my $offer_row = $dbh->selectrow_hashref($sql, undef, ($offer_id, $md5));

	if (!$offer_row) {
		$self->param('post_offer_acceptance_error' => "It appears there is no offer with id $offer_id made to profile $md5 that isnt already completed. I cannot help you.");
		return 0;
	}
	#if validate_offer checks out, then we are good to go (and can proceed).
	if (!$self->validate_offer({offer_row => $offer_row})) {
		$self->param('post_offer_acceptance_error' => "TxManager->post_offer_acceptance: validate_offer failed with this message: " . $self->param('validate_offer_error'));
		return 0;
	} 
	if ($md5 ne $offer_row->{to_md5}) {
		$self->param('post_offer_acceptance_error' => "Hold it there shifty - the offer wasnt even made to you");
		return 0;
	}
	
	#establish timestamp for accept
	my $time = localtime;
	my $accept_time = $time->mysql_datetime;
	
	#ok simply being here means that the offer was valid and so now we can proceed with the acceptance. (thus completing the transaction?)
		#so what is the message to sign?
	my $accept_message = {
		offer_message   => $offer_row->{offer_message},
		offer_signature => $offer_row->{offer_signature},
		accept_time     => $accept_time,
	}; #easy, no? #accept message is just the offer message and the offer signature and the timestamp
	my $accept_messagee_serialzed = freeze($accept_message);

	$self->{wa}->ch_debug(['post_offer_acceptance with to_md5 (person who is the accepting party) of $offer_row->{to_md5}', $offer_row->{to_md5}]);
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self->{wa});
	my $passphrase = $args->{passphrase};
	my $private_key = $prof_man->get_profile_private_key($md5, $passphrase); #dont bother re-validating at this point.
	if (!$private_key) { confess 'ok something went wrong didnt get a private key back from prof_man->get_profile_private_key'; }

	require spApp::RSACrypto;
	my $crypto   = spApp::RSACrypto->new($self->{wa});

	#note signature will just be a block of scalar data, not an object.
	my $signature;
	my $result = eval {
		$signature = $crypto->sign_serialized_data({
			private_key     => $private_key,
			serialized_data => $accept_messagee_serialzed,
		});
		return 1;
	};
	$private_key->hide(); #so we dont have the passphrase floating around in memory (or so i hope)
	if (!$result) {
		$self->param('post_offer_acceptance_error' => "sorry to say, but there was a problem signing the offer-accept message. check private key passphrase.");
		return 0;
	}
		
	#ok it is signed. w00t. 
	#update the offers table.
	my $sql = 'UPDATE offer SET accept_message = ?, accept_signature = ?, accept_time = ? WHERE id = ?';
	my $dbh = $self->{wa}->get_dbh();
	$dbh->do($sql, undef, ($accept_messagee_serialzed, $signature, $accept_time, $offer_id)) or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;

	#Now since I'm going to send a copy of this source code, the templates, and db dump to John, I want to at least make sure that the function I just finished writing actually works before I send him the code.
	my $offer_row = $dbh->selectrow_hashref('SELECT * FROM offer WHERE id = ?', undef, $offer_id);
	my $validate_the_offer_again_just_for_fun = $self->validate_offer({offer_row => $offer_row});
	$self->{wa}->ch_debug(['foospot: validate_offer gave us', $validate_the_offer_again_just_for_fun]);

	my $validate_the_acceptance_just_for_fun  = $self->validate_offer_acceptance({offer_row => $offer_row});
	$self->{wa}->ch_debug(['foospot2: validate_offer_acceptance gave us', $validate_the_acceptance_just_for_fun, $self->param('validate_offer_acceptance_error')]);

	if (!$validate_the_offer_again_just_for_fun || !$validate_the_acceptance_just_for_fun) {
		$self->param('post_offer_acceptance_error' => "Something is very wrong with either the validate_offer or validate_offer_acceptance code.");
		return 0;
	}

	#This is where I have gotten to John. It still doesnt do a lot of things. It doesnt take the accepted offer and make a real transaction record out of it, signed by the STR itself. It doesnt allow external parties to upload their own self-generated profile (and public key) information. It doesnt allow individuals to upload their own signed offer and accept messages. It doesnt have an ability to search offers. It doesnt have any access control at all yet. And I imagine there are tons of other things that could be built in as well to help facilitate people in making transactions securely.

	return 1;
}

sub validate_offer {
	my $self = shift;
	my $args = shift;
	
	my $offer_row     = $args->{offer_row};
	my $offer_message = thaw($offer_row->{offer_message});

	$self->{wa}->ch_debug(['TxManager: validate_offer, $offer_message looks like:', $offer_message, 'args were', $args]);

	#how to validate? 
		#- validate the from_md5 profile
		#- validate the to_md5 profile
		#- valiate the offer_message signature
		#- check that all the fields of the offer_message match up with their db row counterparts.

	$self->param('validate_offer_error' => undef);

	#make sure we got an offer id.
	my $offer_id = $offer_row->{id}; #just for error infos -- but we shouldnt proceed if it is missing.
	if (!$offer_id) {
		$self->param('validate_offer_error' => "offer id was not included in the passed in offer row. we really, really want to get that.");
		return 0;
	}

	#validate the profiles
	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self->{wa});
	if (!$prof_man->validate_profile($offer_row->{from_md5})) {
		$self->param('validate_offer_error' => "offer id $offer_id has errors: offer from_md5 $offer_row->{from_md5} profile appears illegitimate. see self->param(validate_profile_error) for more details.");
		return 0;
	}
	if (!$prof_man->validate_profile($offer_row->{to_md5})) {
		$self->param('validate_offer_error' => "offer id $offer_id has errors: offer to_md5 $offer_row->{to_md5} profile appears illegitimate. see self->param(validate_profile_error) for more details.");
		return 0;
	}

	#validate the offer message signature.
	require spApp::RSACrypto;
	my $crypto   = spApp::RSACrypto->new($self->{wa});
	my $good_offer_sig = $crypto->validate_signature({
		public_key      => $prof_man->get_profile_public_key($offer_row->{from_md5}),
		serialized_data => $offer_row->{offer_message},
		signature       => $offer_row->{offer_signature}
	});
	if (!$good_offer_sig) {
		$self->param('validate_offer_error' => "offer id $offer_id has errors: failed to validate the offer message. signature didnt check out. offer was supposedly made/signed by profile md5 '$offer_message->{from_md5}'.");
		return 0;
		#(well, we could have done the signing as proxy but then there REALLY shouldnt be anything wrong with it unless some joker is effing with the db.)
	}

	#ok so we know that the signature on the offer message is good. But the serialized offer message is not what people are going to be looking at, they'll be looking at row data from the db records. We must ensure that the row data matches what is inside the offer message, otherwise we have a major failure (someone messing with db)
		#make sure that all the data which is inside offer_message mathes the values recorded in the database fields. if anything does not match then we have a failure.
	foreach (keys(%$offer_message)) {
		if ($offer_row->{$_} ne $offer_message->{$_}) {
			$self->param('validate_offer_error' => "offer id $offer_id has errors: offer_message->{$_} does not match the row-explicit $_ field. Is someone tampering with the db?");
			$self->{wa}->ch_debug(['validate_offer - checking the db field match the signed data, db fields were:', $offer_row, 'signed data was', $offer_message]);
			return 0;
		}		
	}
	
	#well, being here means that so far, everything about the offer is good.

#	if ($offer_message->[0] ne $offer_row->{from_md5}) {
#		$self->param('validate_offer_error' => "offer id $offer_id has errors: offer_message->[0] does not match the row-explicit from_md5 field. Is someone tampering with the db?");
#		return 0;
#	}
#	if ($offer_message->[4] ne $offer_row->{to_md5}) {
#		$self->param('validate_offer_error' => "offer id $offer_id has errors: offer_message->[4] does not match the row-explicit to_md5 field. Is someone tampering with the db?");
#		return 0;
#	}
#	if ($offer_message->[5] ne $offer_row->{offer_time}) {
#		$self->param('validate_offer_error' => "offer id $offer_id has errors: offer_message->[5] does not match the row-explicit offer_time field. Is someone tampering with the db?");
#		return 0;
#	}


	return 1; #if we get here, we're good to go.
}

sub validate_offer_acceptance {
	my $self = shift;
	my $args = shift;
	
	my $offer_row = $args->{offer_row};
	my $accept_message = thaw($offer_row->{accept_message});

	#so how in sam hill do we know if the acceptance is valid or invalid? well, we have a very simple method (whoever has this stamp on his hand is invalid!?)
		#check that the accept_message contains an exact copy of the offer_message and an exact copy of the offer_signature
		#verify the signature of the accept_message (using public key of to_md5 profile)
	#if those check out, it has to be good. (right?)	
	$self->param('validate_offer_acceptance_error' => undef);

	#check that the accept_message contains an exact copy of the offer_message and an exact copy of the offer_signature
	#these will be serialized data compares
	my $offer_message_in_row              = $offer_row->{offer_message}; 
	my $offer_message_in_accept_message   = $accept_message->{offer_message};
	if ($offer_message_in_row ne $offer_message_in_accept_message) {
		$self->param('validate_offer_acceptance_error' => "the copy of the offer_message within the accept_message does not match the offer_message in the row. Therefor the offer acceptance appears invalid.");
		return 0;
	}

	#these will be serialized data compares
	my $offer_signature_in_row            = $offer_row->{offer_signature};
	my $offer_signature_in_accept_message = $accept_message->{offer_signature};
	if ($offer_signature_in_row ne $offer_signature_in_accept_message) {
		$self->param('validate_offer_acceptance_error' => "the copy of the offer_signature within the accept_message does not match the offer_signature in the row. Therefor the offer acceptance appears invalid.");
		return 0;
	}

	#verify the signature of the accept_message (using public key of to_md5 profile)
	my $offer_message = thaw($offer_message_in_accept_message);
	my $to_md5 = $offer_message->{to_md5};
	require spApp::ProfileManager;
	require spApp::RSACrypto;
	my $crypto   = spApp::RSACrypto->new($self->{wa});
	my $prof_man = spApp::ProfileManager->new($self->{wa});
	my $good_offer_sig = $crypto->validate_signature({
		public_key      => $prof_man->get_profile_public_key($to_md5),
		serialized_data => $offer_row->{accept_message},
		signature       => $offer_row->{accept_signature}
	});
	if (!$good_offer_sig) {
		$self->param('validate_offer_acceptance_error' => "the accept_signature did not check out. Therefor the offer acceptance appears invalid.");
		return 0;
		#(well, we could have done the signing as proxy but then there REALLY shouldnt be anything wrong with it unless some joker is effing with the db.)
	}
	
	#now theres not a whole bunch of fields to check, really just make sure the accept_time reported in the row matches the accept_time in the accept_message.	
	if ($offer_row->{accept_time} ne $accept_message->{accept_time}) {
		$self->param('validate_offer_error' => "offer id $offer_row->{id} has errors: accept_message->{accept_time} does not match the row-explicit accept_time field. Is someone tampering with the db?");
		return 0;
	}		

	
	#ok we got here so we're good to go - the offer accept must be valid!
	return 1;
}

sub find_profile_accepted_offers {
	my $self = shift;
	
	my $args = shift;
	return $self->find_offers({to_md5 => $args->{md5}, accepted => 'IS NOT NULL'});
}
sub find_profile_posted_offers {
	my $self = shift;
	
	my $args = shift;
	return $self->find_offers({from_md5 => $args->{md5}, accepted => 'IS NOT NULL'});
}

sub find_offers {
	my $self = shift;
	
	#this is to return data structures that can be used for output.
	
	my $args = shift;
	my $to_md5 = $args->{to_md5};
	my $from_md5 = $args->{from_md5};
	my $include_name_email_of = $args->{include_name_email_of};
	
	my $accepted_search = $args->{accepted};
	my $md5_search_blk = []; my $bnd = []; #searchblocks and bindparams
	if ($to_md5) { push(@$md5_search_blk, 'to_md5 = ?'); push(@$bnd, $to_md5); }
	if ($from_md5) { push(@$md5_search_blk, 'from_md5 = ?'); push(@$bnd, $from_md5); }
	my $md5_search = '(' . join(' OR ', @$md5_search_blk) . ')';

	if (!$accepted_search) {
		$accepted_search = 'IS NULL';
	}
	
	my $sql = "SELECT * FROM offer WHERE $md5_search AND accept_time $accepted_search";
	my $dbh = $self->{wa}->get_dbh();
	$self->{wa}->ch_debug(['find_offers: gonna do this sql', $sql, 'with this bind param', $bnd]);
	my $sth = $dbh->prepare($sql);
	$sth->execute(@$bnd) or die "Db error: (with sql -->$sql<--)" . $dbh->errstr;

	require spApp::ProfileManager;
	my $prof_man = spApp::ProfileManager->new($self->{wa}); #used to get the details of the offering-party.

	my $list = [];
	while (my $row = $sth->fetchrow_hashref) {
		if (!$self->validate_offer({offer_row => $row})) {
			$self->{wa}->ch_debug(["find_offers: found an invalid offer row (id is) : $row->{id}", $self->param('validate_offer_error')]);
			next;
		}
		if ($row->{accepted_time} && !$self->validate_offer_acceptance({offer_row => $row})) {
			#we'd not get to this spot unless the row has an accepted time, and we'd not get rows with accepted times unless we we're searching for them. so if it has it, validate it!!
			$self->{wa}->ch_debug(["find_offers: found an invalidly accepted offer row (id is) : $row->{id}", $self->param('validate_offer_acceptance_error')]);
			next;
		}

		#offer checks out (cuz we're here).
		#we can get some details if we've been asked to.
		if ($include_name_email_of) {
			my $profile_info = $prof_man->find_profiles({md5 => $row->{$include_name_email_of}});
			if (!$profile_info) {
				die "umm .. tried to get a validated profile for the offer from $row->{from_md5} -- but that didnt work out. shouldnt see this message";
				next; #just skip it if somehow it was bad.
			}
			$row->{name_email} = $profile_info->{name_email};
		}
		
		#$self->{wa}->ch_debug(['find_offers: found a good one, and tried to get info about who offered: ', $offeror_details]);
		push(@$list, $row);
	}
	
	return $list;
	

}


sub sum_amounts {
	my $self = shift;
	my $args = shift;

	my $completed_offers = $args->{offer_list};
	my $currency_totals = {};
	foreach (@$completed_offers) {
		$currency_totals->{$_->{currency_unit}} += $_->{amount};
	}
	return $currency_totals;
}

sub param {
	my $self = shift;
	return $self->{wa}->param(@_);
}


	
1;
